#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include <glib-object.h>

#include "ws.h"
#include "wsint.h"

#include <X11/extensions/shape.h>

static WsWindow *
window_new (Ws *ws,
	    Window xwindow)
{
    WsWindow *window = g_new0 (WsWindow, 1);

    g_assert (!g_hash_table_lookup (ws->xresources, (gpointer)xwindow));
    
    _ws_drawable_init (&window->drawable, ws, xwindow);

    g_assert (xwindow == window->drawable.xid);
    
    ws_begin_error_trap (ws);
    
    XSelectInput (ws->xdisplay, xwindow,
		  StructureNotifyMask);

    ws_end_error_trap (ws);
    
    return window;
}

WsWindow *
_ws_window_ensure (Ws	  *ws,
		   Window  xwindow)
{
    WsWindow *window;

    if (!xwindow)
	return NULL;
    
    window = g_hash_table_lookup (ws->xresources, (gpointer)xwindow);
    
    if (!window)
	window = window_new (ws, xwindow);

    return window;
}

void
ws_window_set_input_shape	  (WsWindow    *window,
				   WsRegion    *shape)
{
    Display *xdisplay = window->drawable.ws->xdisplay;
    
    XFixesSetWindowShapeRegion (xdisplay, window->drawable.xid, ShapeInput,
				0, 0, shape->xregion);
}

/* Create a window and make it the active GL window
 * The right way forward here is most likely to have a new
 * subclass "GlWindow" with methods
 *
 *	   begin()				// make it active if it isn't
 *	   end()				// 
 *         swap_buffers();			// 
 *
 */

WsWindow *
ws_window_new_gl (WsWindow *parent)
{
    int attrib[] = { GLX_RGBA,
		     GLX_RED_SIZE, 1,
		     GLX_GREEN_SIZE, 1,
		     GLX_BLUE_SIZE, 1,
		     GLX_STENCIL_SIZE, 1,
		     GLX_DOUBLEBUFFER,
		     GLX_DEPTH_SIZE, 1,
		     None };
    XVisualInfo *visinfo;
    GLXContext context;
    XSetWindowAttributes attr;
    Window xwindow;
    WsScreen *screen;
    Display *xdisplay;
    
    g_return_val_if_fail (parent != NULL, NULL);
    
    xdisplay = parent->drawable.ws->xdisplay;
    
    screen = ws_drawable_query_screen ((WsDrawable *)parent);
    
    visinfo = glXChooseVisual (xdisplay, ws_screen_get_number (screen), attrib);
    
    /* FIXME: report an error instead of asserting */
    g_assert (visinfo);
    
    context = glXCreateContext (xdisplay, visinfo, NULL, True);
    
    /* FIXME: report an error instead of asserting */
    g_assert (context);
    
    attr.colormap = XCreateColormap (
	xdisplay, parent->drawable.xid, visinfo->visual, AllocNone);

    xwindow = XCreateWindow (xdisplay, parent->drawable.xid,
#if 0
			     400, 400, 1200, 800, 0,
#endif
			     0, 0,
			     screen->xscreen->width,
			     screen->xscreen->height, 0,
			     visinfo->depth, InputOutput, visinfo->visual,
			     CWColormap, &attr);
    
    glXMakeCurrent (xdisplay, xwindow, context);
    
    XSelectInput (xdisplay, xwindow, ExposureMask | PointerMotionMask);

    WsRegion *region = ws_region_new (parent->drawable.ws);
    WsWindow *result;
    
    result = _ws_window_ensure (parent->drawable.ws, xwindow);
    
    ws_window_set_input_shape (result, region);

    ws_region_unref (region);

    return result;
    
}

void
ws_window_redirect_subwindows (WsWindow *window)
{
    g_return_if_fail (window->drawable.ws->composite.available);
    
    XCompositeRedirectSubwindows (
	window->drawable.ws->xdisplay, window->drawable.xid,
	CompositeRedirectManual);
}

void
ws_window_map (WsWindow *window)
{
    Display *xdisplay;
    
    g_return_if_fail (window != NULL);
    
    xdisplay = window->drawable.ws->xdisplay;
    
    XMapWindow (xdisplay, window->drawable.xid);
}

void
ws_window_raise (WsWindow *window)
{
    g_return_if_fail (window != NULL);
    
    XRaiseWindow (window->drawable.ws->xdisplay, window->drawable.xid);
}

void
ws_window_lower (WsWindow *window)
{
    g_return_if_fail (window != NULL);

    XLowerWindow (window->drawable.ws->xdisplay, window->drawable.xid);
}

void
ws_window_gl_swap_buffers (WsWindow *window)
{
    glXSwapBuffers (window->drawable.ws->xdisplay, window->drawable.xid);
}

void
ws_window_unredirect (WsWindow *window)
{
    XCompositeUnredirectWindow (
	window->drawable.ws->xdisplay,
	window->drawable.xid, CompositeRedirectManual);
}

void
ws_window_set_override_redirect (WsWindow *window, gboolean override_redirect)
{
    XSetWindowAttributes attrs;
    
    attrs.override_redirect = !!override_redirect;
    
    XChangeWindowAttributes (window->drawable.ws->xdisplay, window->drawable.xid,
			     CWOverrideRedirect, &attrs);
}

gboolean
ws_window_query_input_only (WsWindow *window)
{
    XWindowAttributes attrs;

    if (!XGetWindowAttributes (window->drawable.ws->xdisplay, window->drawable.xid, &attrs))
	return FALSE;

    return attrs.class == InputOnly;
}

gboolean
ws_window_query_mapped (WsWindow *window)
{
    XWindowAttributes attrs;

    if (!XGetWindowAttributes (window->drawable.ws->xdisplay, window->drawable.xid, &attrs))
	return FALSE;

    return attrs.map_state == IsUnviewable || attrs.map_state == IsViewable;
}

gboolean
ws_window_query_viewable (WsWindow *window)
{
    XWindowAttributes attrs;

    if (!XGetWindowAttributes (window->drawable.ws->xdisplay, window->drawable.xid, &attrs))
	return FALSE;

    return attrs.map_state == IsViewable;
}

WsFormat
ws_window_query_format (WsWindow *window)
{
    XWindowAttributes attrs;
    
    XGetWindowAttributes (window->drawable.ws->xdisplay, window->drawable.xid, &attrs);

    if (attrs.depth == 16 &&
	attrs.visual->red_mask == 0xf800 &&
	attrs.visual->green_mask == 0x7e0 &&
	attrs.visual->blue_mask == 0x1f)
    {
	return WS_FORMAT_RGB_16;
    }
    else if (attrs.depth == 24 &&
	     attrs.visual->red_mask == 0xff0000 &&
	     attrs.visual->green_mask == 0xff00 &&
	     attrs.visual->blue_mask == 0xff)
    {
	return WS_FORMAT_RGB_24;
    }
    else if (attrs.depth == 32 &&
	     attrs.visual->red_mask == 0xff0000 &&
	     attrs.visual->green_mask == 0xff00 &&
	     attrs.visual->blue_mask == 0xff)
    {
	return WS_FORMAT_ARGB_32;
    }
    else
    {
	g_warning ("Unknown visual format depth=%d, r=%#lx/g=%#lx/b=%#lx",
		   attrs.depth, attrs.visual->red_mask, attrs.visual->green_mask, attrs.visual->blue_mask);

	return WS_FORMAT_UNKNOWN;
    }
}

GList *
ws_window_query_subwindows (WsWindow *window)
{
    Window root, parent;		/* dummies */
    Window *children;
    guint n_children;
    int i;
    GList *result = NULL;
    
    if (XQueryTree (window->drawable.ws->xdisplay,
		    window->drawable.xid, &root, &parent,
		    &children, &n_children))
    {
	for (i = 0; i < n_children; ++i)
	{
	    WsWindow *child = _ws_window_ensure (window->drawable.ws,
						 children[i]);
	    
	    result = g_list_prepend (result, child);
	}

	XFree (children);
    }
    
    return result;
}

/* FIXME: This function just uses XFetchName - it should do the
 * _NET_WM stuff. If this function should exist at all, which maybe it
 * shouldn't
 */

gchar *
ws_window_query_title (WsWindow *window)
{
    char *result = NULL;
    char *name;
    
    if (XFetchName (window->drawable.ws->xdisplay,
		    window->drawable.xid, &name))
    {
	result = g_strdup (name);
	XFree (name);
    }
    else
    {
	result = g_strdup ("No title");
    }

    return result;
}

WsWindow *
ws_window_lookup (Ws     *ws,
		  Window  xwindow)
{
    return _ws_window_ensure (ws, xwindow);
}


WsPixmap *
ws_window_name_pixmap (WsWindow *window)
{
    Pixmap pixmap = XCompositeNameWindowPixmap (window->drawable.ws->xdisplay,
						window->drawable.xid);
    
    return _ws_pixmap_ensure (window->drawable.ws, pixmap);
}

void
ws_window_set_configure_callback (WsWindow   *window,
				  WsConfigureCallback cb,
				  gpointer	data)
{
    window->configure_callback = cb;
    window->configure_data = data;
}

void
_ws_window_process_event (WsWindow   *window,
			  XEvent     *xevent)
{
    if (xevent->type == ConfigureNotify)
    {
	XConfigureEvent *configure = &xevent->xconfigure;
	WsWindow *above;

	above = _ws_window_ensure (window->drawable.ws, configure->above);
	
	if (window->configure_callback)
	{
	    window->configure_callback (window,
					configure->x, configure->y,
					configure->width, configure->height,
					configure->border_width,
					above,
					configure->override_redirect,
					window->configure_data);
	}
    }
}
